""" Deletion Views """

__all__ = (
    "EntityMassDeletion",
)

import json

from django.core.exceptions import PermissionDenied
from django.db import models
from django.http import (
    Http404, HttpResponse, HttpResponseGone, HttpResponsePermanentRedirect,
    JsonResponse
)
from django.urls import reverse
from django.utils import timezone
from django.views.generic.base import TemplateView

from spkcspider.apps.spider.queryfilters import loggedin_active_tprotections_q

from ..models import AssignedContent, UserComponent
from ._core import UCTestMixin


class EntityMassDeletion(UCTestMixin, TemplateView):
    """
        Delete Contents and Usercomponents
    """
    preserved_GET_parameters = {"token", "search", "id", "cid", "uc"}
    own_marked_for_deletion = False
    template_name = "spider_base/entity_mass_deletion.html"

    @staticmethod
    def to_int(value):
        try:
            return int(value)
        except ValueError:
            return None

    def get_context_data(self, **kwargs):
        kwargs["selected_components"] = set(self.request.GET.getlist("uc"))
        kwargs["selected_contents"] = set(map(
            self.to_int, self.request.GET.getlist("cid")
        ))
        return super().get_context_data(**kwargs)

    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except PermissionDenied:
            raise Http404()

    def test_func(self):
        return self.has_special_access(
            user_by_login=True, user_by_token=True,
            staff=False, superuser=False
        )

    def get(self, request, *args, **kwargs):
        now = timezone.now()
        user = self.usercomponent.user
        travel = self.get_travel_for_request().filter(
            loggedin_active_tprotections_q
        )

        travel_contents_q = (
            models.Q(travel_protected__in=travel) |
            models.Q(usercomponent__travel_protected__in=travel)
        )

        components = {}

        if not self.usercomponent.is_index:
            component_query = UserComponent.objects.filter(
                id=self.usercomponent.id
            )
        else:
            component_query = UserComponent.objects.filter(
                user=user
            )

        content_query = AssignedContent.objects.filter(
            usercomponent__user=user
        )

        ignored_content_ids = frozenset(
            content_query.filter(
                travel_contents_q
            ).values_list("id", flat=True)
        )
        ignored_component_ids = frozenset(
            component_query.filter(
                travel_protected__in=travel
            ).values_list("id", flat=True)
        )

        for uc in component_query:
            item = {
                "ob": uc,
                "contents": {},
                "deletion_active": uc.deletion_requested is not None
            }
            item["deletion_date"] = \
                self.calculate_deletion_component(
                    uc,
                    now,
                    ignored_content_ids,
                    log=item["contents"],
                    del_expired=True
                )

            if item["deletion_date"] is None:
                if uc == self.usercomponent:
                    self.own_marked_for_deletion = True
                continue
            if uc.id not in ignored_component_ids:
                components[uc.name] = item
        return self.render_to_response(self.get_context_data(
            hierarchy=components,
            now=now,
            ignored_component_ids=ignored_component_ids
        ))

    def post(self, request, *args, **kwargs):
        now = timezone.now()
        user = self.usercomponent.user
        if request.content_type == "application/json":
            json_data = json.loads(request.body)
            delete_components = json_data.get("delete_components", [])
            reset_components = json_data.get("reset_components", [])
            delete_contents = json_data.get("delete_contents", [])
            reset_contents = json_data.get("reset_contents", [])
        else:
            delete_components = request.POST.getlist("delete_components")
            reset_components = request.POST.getlist("reset_components")
            delete_contents = request.POST.getlist("delete_contents")
            reset_contents = request.POST.getlist("reset_contents")
        try:
            reset_components = set(reset_components)
            delete_components = set(delete_components)
            delete_components.difference_update(reset_components)
            reset_contents = set(map(int, reset_contents))
            delete_contents = set(map(
                int, delete_contents
            ))
            delete_contents.difference_update(reset_contents)

        except Exception:
            return HttpResponse(status=400)

        travel = self.get_travel_for_request().filter(
            loggedin_active_tprotections_q
        )

        travel_contents_q = (
            models.Q(travel_protected__in=travel) |
            models.Q(usercomponent__travel_protected__in=travel)
        )

        components = {}

        if not self.usercomponent.is_index:
            component_query = UserComponent.objects.filter(
                id=self.usercomponent.id
            )
        else:
            component_query = UserComponent.objects.filter(
                user=user
            )

        content_query = AssignedContent.objects.filter(
            usercomponent__user=user
        )

        ignored_content_ids = frozenset(
            content_query.filter(
                travel_contents_q
            ).values_list("id", flat=True)
        )
        ignored_component_ids = frozenset(
            component_query.filter(
                travel_protected__in=travel
            ).values_list("id", flat=True)
        )

        delete_contents.difference_update(ignored_content_ids)
        reset_contents.difference_update(ignored_content_ids)

        component_query.exclude(id__in=ignored_component_ids).filter(
            name__in=delete_components,
            deletion_requested__isnull=True
        ).update(
            deletion_requested=now
        )
        component_query.exclude(id__in=ignored_component_ids).filter(
            name__in=reset_components
        ).update(
            deletion_requested=None
        )
        content_query.exclude(
            id__in=ignored_content_ids
        ).filter(
            id__in=delete_contents,
            deletion_requested__isnull=True
        ).update(
            deletion_requested=now
        )
        content_query.exclude(
            id__in=ignored_content_ids
        ).filter(id__in=reset_contents).update(
            deletion_requested=None
        )

        for uc in component_query:
            item = {
                "ob": uc,
                "contents": {},
                "deletion_active":
                    (
                        uc.name in delete_components or
                        uc.deletion_requested
                    ) and not uc.is_index
            }

            item["deletion_date"] = \
                self.calculate_deletion_component(
                    uc, now, ignored_content_ids, item["contents"],
                    del_expired=True
                )
            if item["deletion_date"] is None:
                if uc == self.usercomponent:
                    self.own_marked_for_deletion = True
                continue

            if uc.id not in ignored_component_ids:
                components[uc.name] = item

        return self.render_to_response(self.get_context_data(
            hierarchy=components,
            now=now,
            ignored_component_ids=ignored_component_ids
        ))

    def options(self, request, *args, **kwargs):
        ret = super().options(request, *args, **kwargs)
        ret["Access-Control-Allow-Origin"] = self.request.get_host()
        ret["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS"
        return ret

    def render_to_response(self, context):
        if not self.own_marked_for_deletion:
            if "application/json" in (
                self.request.headers.get("accept"), self.request.content_type
            ):
                # last time it is used
                for component_val in context["hierarchy"].values():
                    component_val["name"] = component_val["ob"].name
                    component_val["description"] = \
                        component_val["ob"].description
                    component_val.pop("ob")
                    for content_val in component_val["contents"].values():
                        content_val["name"] = content_val["ob"].name
                        content_val["description"] = \
                            content_val["ob"].description
                        content_val.pop("ob")
                response = JsonResponse(context["hierarchy"])
            else:
                response = super().render_to_response(context)
        elif self.request.user.is_authenticated:
            # last resort: redirect to index
            index = self.request.user.usercomponent_set.get("index")
            response = HttpResponsePermanentRedirect(
                "{}?{}".format(
                    reverse(
                        'spider_base:entity-delete',
                        kwargs={"token": index.token}
                    ), context["sanitized_GET"]
                )
            )
        else:
            response = HttpResponseGone()
        response["Access-Control-Allow-Origin"] = self.request.get_host()
        response["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS"
        return response



import datetime

import requests

from django.conf import settings
from django.core import exceptions
from django.db import models
from django.test import Client
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from spider_domainauth.abstract_models import BaseReverseToken
from spkcspider.constants.verifier import VERIFICATION_CHOICES

from .conf import get_requests_params
from .functions import get_anchor_domain


def dv_path(instance, filename):
    return 'dvfiles/{}/{}.{}'.format(
        datetime.datetime.now().strftime("%Y/%m"),
        instance.hash,
        "ttl"
    )


class VerifySourceObject(BaseReverseToken):
    id = models.BigAutoField(primary_key=True, editable=False)
    url = models.URLField(
        max_length=(
            400 if (
                settings.DATABASES["default"]["ENGINE"] !=
                "django.db.backends.mysql"
            ) else 255
        ), db_index=True, unique=True
    )
    get_params = models.TextField()

    def get_url(self, access=None):
        if access:
            split = self.url.rsplit("view", 1)
            if len(split) == 1:
                raise Exception()
            return "{}{}/?{}".format(split[0], access, self.get_params)
        return "{}?{}".format(self.url, self.get_params)


class DataVerificationTag(models.Model):
    """ Contains verified data """
    # warning: never depend directly on user, seperate for multi-db setups
    id = models.BigAutoField(primary_key=True, editable=False)
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)

    hash = models.SlugField(
        unique=True, db_index=True, null=False,
        max_length=255
    )
    dvfile = models.FileField(
        upload_to=dv_path, null=True, blank=True, help_text=_(
            "File with data to verify"
        )
    )
    source = models.ForeignKey(
        VerifySourceObject, null=True, blank=True, on_delete=models.CASCADE
    )
    # url = models.URLField(max_length=600)
    data_type = models.CharField(default="layout", max_length=20)
    checked = models.DateTimeField(null=True, blank=True)
    verification_state = models.CharField(
        default="pending",
        max_length=10, choices=VERIFICATION_CHOICES
    )
    note = models.TextField(default="", blank=True)

    class Meta:
        permissions = [("can_verify", "Can verify Data Tag?")]

    def __str__(self):
        return "DVTag: ...%s" % self.hash[:30]

    def get_absolute_url(self):
        return reverse(
            "spider_verifier:verify",
            kwargs={
                "hash": self.hash
            }
        )

    def callback(self, hostpart=None):
        if not hostpart:
            hostpart = get_anchor_domain()
        if self.source and self.data_type.endswith("_cb"):
            vurl = self.source.get_url("verify")
            body = {
                "url": "{}{}".format(
                    hostpart, self.get_absolute_url()
                )
            }
            params, inline_domain = get_requests_params(vurl)

            if inline_domain:
                resp = Client().post(
                    vurl, follow=True, secure=True, data=body,
                    Connection="close", SERVER_NAME=inline_domain
                )
                if resp.status_code != 200:
                    raise exceptions.ValidationError(
                        _("Retrieval failed: %(reason)s"),
                        params={"reason": resp.content},
                        code="error_code:{}".format(resp.status_code)
                    )
            else:
                try:
                    with requests.post(
                        vurl, data=body, headers={"Connection": "close"},
                        **params
                    ) as resp:
                        if resp.status_code != 200:
                            raise exceptions.ValidationError(
                                _("Retrieval failed: %(reason)s"),
                                params={"reason": resp.reason},
                                code="error_code:{}".format(resp.status_code)
                            )
                except requests.exceptions.Timeout:
                    raise exceptions.ValidationError(
                        _('url timed out: %(url)s'),
                        params={"url": vurl},
                        code="timeout_url"
                    )
                except requests.exceptions.ConnectionError:
                    raise exceptions.ValidationError(
                        _('invalid url: %(url)s'),
                        params={"url": vurl},
                        code="invalid_url"
                    )



# Generated by Django 3.0.1 on 2019-12-30 13:05

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0011_auto_20191230_1305'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='travelprotection',
            name='protect_components',
        ),
        migrations.RemoveField(
            model_name='travelprotection',
            name='protect_contents',
        ),
        migrations.DeleteModel(
            name='TravelProtection',
        ),
        migrations.DeleteModel(
            name='LinkContent',
        ),
        migrations.RemoveField(
            model_name='assignedcontent',
            name='object_id',
        ),
        migrations.CreateModel(
            name='LinkContent',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('spider_base.datacontent',),
        ),
        migrations.CreateModel(
            name='TravelProtection',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('spider_base.datacontent',),
        ),
        migrations.AlterModelOptions(
            name='travelprotection',
            options={'permissions': [('use_dangerous_travelprotections', 'Can use dangerous TravelProtections')]},
        ),
    ]




import logging
from urllib.parse import urljoin

from cryptography import exceptions
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.x509 import load_pem_x509_certificate

from django.conf import settings
from django.db import models
from django.http import HttpResponsePermanentRedirect, JsonResponse
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.translation import pgettext
from spkcspider.apps.spider.conf import get_anchor_domain, get_anchor_scheme
from spkcspider.apps.spider import registry
from spkcspider.apps.spider.models import DataContent
from spkcspider.constants import VariantType
from spkcspider.utils.security import get_hashob
from spkcspider.utils.fields import add_by_field

logger = logging.getLogger(__name__)


# Create your models here.

ID_VERIFIERS = {

}


@add_by_field(registry.contents, "_meta.model_name")
class PublicKey(DataContent):
    expose_name = False
    expose_description = True
    appearances = [
        {"name": "PublicKey", "ctype": VariantType.unique}
    ]

    class Meta:
        proxy = True

    @classmethod
    def localize_type_name(cls, name):
        _ = pgettext
        return _("content name", "Public Key")

    def get_info(self):
        ret = super().get_info()
        h = get_hashob()
        if self.prepared_attachements:
            key = self.prepared_attachements["attachedblobs"][0]
        else:
            key = self.associated.attachedblobs.get(name="key")
        h.update(key.blob)
        pubkeyhash = ""
        k = self.get_key_ob(key.blob)
        if k:
            pem = k.public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo
            )
            h2 = get_hashob()
            h2.update(pem)
            pubkeyhash = "pubkeyhash=%s=%s\x1e" % (
                settings.SPIDER_HASH_ALGORITHM.name, h2.finalize().hex()
            )
        return "%shash=%s=%s\x1e%s" % (
            ret, settings.SPIDER_HASH_ALGORITHM.name, h.finalize().hex(),
            pubkeyhash
        )

    def get_key_ob(self, key=None):
        if not key:
            key = self.associated.attachedblobs.get(name="key").blob
        try:
            if b"-----BEGIN CERTIFICATE-----" in key:
                pubkey = load_pem_x509_certificate(
                    key, default_backend()
                ).public_key()
            else:
                pubkey = serialization.load_pem_public_key(
                    key, default_backend()
                )
            return pubkey
        except exceptions.UnsupportedAlgorithm:
            return None
        except ValueError:
            pass

    def get_key_name(self):
        if self.prepared_attachements:
            key = self.prepared_attachements["attachedblobs"][0].as_bytes
        else:
            key = self.associated.attachedblobs.get(name="key").as_bytes
        key = key.decode("ascii")
        # ssh key
        split = key.rsplit(" ", 1)
        if len(split) == 2 and "@" in split[1]:
            return split
        # other key, e.g PEM
        return (key, None)

    def get_content_name(self):
        name = self.get_key_name()[1]
        if not name:
            name = "{}...".format(
                self.associated.getlist("hash", amount=1)[0].split(
                    "=", 1
                )[-1][:20]
            )
        if len(self.associated.description) > 0:
            name = "{}: {}".format(name, self.associated.description[:20])
        return "Key: {}".format(name)

    def get_form(self, scope):
        from .forms import KeyForm
        return KeyForm

    def access_view(self, **kwargs):
        kwargs["object"] = self
        kwargs["key"] = \
            self.associated.attachedblobs.get(
                name="key"
            ).as_bytes.decode("ascii")
        kwargs["hash_algo"] = settings.SPIDER_HASH_ALGORITHM.name
        kwargs["hash"] = self.associated.getlist(
            "hash=%s" % settings.SPIDER_HASH_ALGORITHM.name, 1
        )[0]
        k = self.associated.getlist(
            "pubkeyhash=%s" % settings.SPIDER_HASH_ALGORITHM.name, 1
        )
        if k:
            kwargs["pubkeyhash"] = k[0]
        return render_to_string(
            "spider_keys/key.html", request=kwargs["request"],
            context=kwargs
        )

# Anchors can ONLY be used for server content.
# Clients have to be verified seperately
# 1 To verify client use this trick:
# requesting party redirects user to server, with a return address in GET
# the server generates a secret
# the user authenticates and redirects to the url plus hash of secret
# the server sends secret via post to requesting party
# the requesting party can verify that both: client and server are connected
# this means: anchors are valid for this client
# or 2: generate a token
# or 3: use oauth

# advantage of 1,2: simple
# implementation: if requester in GET, create token and
# redirect after auth to requester with hash of token
# post token to requester

# TODO: make hash algorithm configurable


class AnchorMixing(object):
    expose_name = False
    expose_description = True

    def get_abilities(self, context):
        return {"anchor"}

    def get_priority(self):
        return -10

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret.setdefault("initial", {})
        ret["initial"]["identifier"] = self.get_identifier(kwargs["request"])
        ret["scope"] = kwargs["scope"]
        return ret

    def get_identifier(self, request):
        """ returns id of content, server """
        # security: id can only be faked by own server
        # this should never happen, except with admin access to server
        if not self.associated_id:
            return None
        ret = urljoin(
            "{}://{}".format(
                get_anchor_scheme(), get_anchor_domain()
            ), reverse(
                "spider_keys:anchor-permanent",
                kwargs={"pk": self.associated_id}
            )
        )
        return ret


@add_by_field(registry.contents, "_meta.model_name")
class AnchorServer(AnchorMixing, DataContent):
    """ identify by server """
    appearances = [
        {
            "name": "AnchorServer",
            "ctype": VariantType.anchor,
            "strength": 0
        }
    ]

    class Meta:
        proxy = True

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", "Server-based Anchor")

    def get_form(self, scope):
        from .forms import AnchorServerForm
        return AnchorServerForm

    def get_content_name(self):
        return "Anchor: {}".format(self.associated_id)

    def access_anchor(self, **kwargs):
        if self.new_url:
            ret = HttpResponsePermanentRedirect(redirect_to=self.new_url)
        else:
            ret = JsonResponse(self.old_urls)
        ret["Access-Control-Allow-Origin"] = "*"
        return ret


@add_by_field(registry.contents, "_meta.model_name")
class AnchorKey(AnchorMixing, DataContent):
    """ domain name of pc, signed """
    expose_name = False
    expose_description = True
    appearances = [
        {
            "name": "AnchorKey",
            "ctype": VariantType.anchor+VariantType.unique,
            "strength": 0
        }
    ]

    class Meta:
        proxy = True

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", "Key-based Anchor")

    def get_content_name(self):
        name = self.associated.attached_to_content.content.get_key_name()[1]
        if not name:
            name = "{}...".format(
                self.associated.attached_to_content.getlist(
                    "hash", amount=1
                )[0].split(
                    "=", 1
                )[-1][:20]
            )
        if len(self.associated.description) > 0:
            name = "{}: {}".format(name, self.associated.description[:20])
        return "AnchorKey: {}".format(name)

    def get_form(self, scope):
        from .forms import AnchorKeyForm
        return AnchorKeyForm

    def access_anchor(self, **kwargs):
        ret = JsonResponse([])
        ret["Access-Control-Allow-Origin"] = "*"
        return ret

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret.setdefault("initial", {})
        ret["request"] = kwargs["request"]
        ret["scope"] = kwargs["scope"]
        return ret

    def get_info(self):
        ret = super().get_info()
        k = self.associated.attached_to_content.getlist(
            "pubkeyhash", amount=1, fullkey=True
        )
        return "%s%s\x1e%s\x1e" % (
            ret,
            self.associated.attached_to_content.getlist(
                "hash", amount=1, fullkey=True
            )[0],
            k[0] if k else ""
        )


# TODO: implement
#
class AnchorLink(AnchorMixing):
    """
        Anchor by Organisation, e.g. government,
        verifier returns token pointing to url
    """

    verified_by = models.URLField(max_length=400)

    appearances = [
        {
            "name": "AnchorLink",
            "ctype": VariantType.anchor + VariantType.unique,
            "strength": 10
        }
    ]

    def access_anchor(self, **kwargs):
        ret = JsonResponse(self.verified_by)
        ret["Access-Control-Allow-Origin"] = "*"
        return ret

    def get_form(self, scope):
        raise NotImplementedError

    def get_info(self):
        return "{}verified_by={}\x1e".format(
            super().get_info(unique=True, unlisted=False),
            self.verified_by
        )



__all__ = ["get_requests_params"]


from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from spkcspider.constants import host_tld_matcher
from spkcspider.utils.settings import get_settings_func


def get_requests_params(url):
    _url = host_tld_matcher.match(url)
    if not _url:
        raise ValidationError(
            _("Invalid URL: \"%(url)s\""),
            code="invalid_url",
            params={"url": url}
        )
    _url = _url.groupdict()
    mapper = getattr(
        settings, "VERIFIER_REQUEST_KWARGS_MAP",
        settings.SPIDER_REQUEST_KWARGS_MAP
    )
    return (
        mapper.get(
            _url["host"],
            mapper.get(
                _url["tld"],  # maybe None but then fall to retrieval 3
                mapper[b"default"]
            )
        ),
        get_settings_func(
            "VERIFIER_INLINE", "SPIDER_INLINE",
            "spkcspider.apps.spider.functions.clean_spider_inline",
            exclude=frozenset({True})
        )(_url["host"])
    )




__all__ = ("BaseContent",)

import logging
from datetime import timedelta
from urllib.parse import urljoin

from rdflib import RDF, XSD, BNode, Graph, Literal, URIRef

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.core.files.base import File
from django.db import models, transaction
from django.http import Http404
from django.http.response import HttpResponse, HttpResponseBase
from django.middleware.csrf import CsrfViewMiddleware
from django.template.loader import render_to_string
from django.utils.html import escape
from django.utils.translation import gettext, pgettext
from django.views.decorators.csrf import csrf_exempt
from spkcspider.constants import VariantType, spkcgraph
from spkcspider.utils.fields import add_property, literalize
from spkcspider.utils.security import create_b64_id_token
from spkcspider.utils.settings import get_settings_func
from spkcspider.utils.urls import merge_get_url

from ..conf import get_anchor_domain
from ..serializing import paginate_stream, serialize_stream

logger = logging.getLogger(__name__)


# don't spam set objects
_empty_set = frozenset()

_blacklisted = set(getattr(
    settings, "SPIDER_BLACKLISTED_MODULES", _empty_set
))

default_abilities = frozenset(
    {"add", "view", "update", "export", "list", "raw", "raw_update"}
)

# never use these names
forbidden_names = {"Content", "UserComponent"}


class BaseContent(models.Model):
    # consider not writing admin wrapper for (sensitive) inherited content
    # this way content could be protected to be only visible to admin, user
    # and legitimated users (if not index)

    # iterable or callable returning iterable containing dicts
    # required keys: name
    # optional: strength (default=0), ctype (default="")
    # conditional required: valid_feature_for
    # max name length is 50 chars
    # max ctype length is 10 chars (10 attributes)
    # valid_feature_for is  a list or if "*" also an string.
    #  "*" inverts choices so parameters are blacklisted instead whitelisted

    # use case: addons for usercontent, e.g. dependencies on external libraries
    # use case: model with different abilities
    appearances = None

    # autofilled
    attached_names = set()

    id: int = models.BigAutoField(primary_key=True, editable=False)
    # every content can specify its own deletion period
    deletion_period = getattr(
        settings, "SPIDER_CONTENTS_DEFAULT_DELETION_PERIOD", timedelta(0)
    )
    # use usercomponent in form instead
    associated = models.OneToOneField(
        "spider_base.AssignedContent", related_name="+",
        on_delete=models.CASCADE, null=True
    )
    # None or dict (key=related_name)
    prepared_attachements = None

    # user can set name
    # if set to "force" name will be enforced
    expose_name = True
    # user can set description
    expose_description = False
    # use if you want to force a token size
    force_token_size = None

    associated_errors = None

    # extra size which should be used in calc
    extra_size = 0

    _content_is_cleaned = False

    class Meta:
        abstract = True
        default_permissions = ()

    def __str__(self):
        if not self.id:
            return self.localize_name(self.associated.ctype.name)
        else:
            return self.associated.name

    def __repr__(self):
        return "<Content: ({}: {})>".format(
            self.associated.usercomponent.username, self.__str__()
        )

    @classmethod
    def static_create(
        cls, *, associated=None, associated_kwargs=None, content_kwargs=None,
        token_size=None
    ):
        if content_kwargs:
            ob = cls(**content_kwargs)
        else:
            ob = cls()
        if associated:
            ob.associated = associated
        else:
            from ..models import AssignedContent
            if not associated_kwargs:
                associated_kwargs = {}
            ob = cls(associated=AssignedContent(**associated_kwargs))
            # abuse cached_property mechanic
            ob.associated.__dict__["content"] = ob
        if token_size is not None:
            ob.associated.token_generate_new_size = token_size
        return ob

    @classmethod
    def localize_name(cls, name):
        """
            Localize type names
        """
        return pgettext("content name", name)

    def get_size(self, prepared_attachements=None):
        from ..models.content_extended import BaseAttached
        # 255 = length name, ignore potential utf8 encoding differences
        s = 255
        if self.expose_description:
            s += len(self.associated.description)
        if prepared_attachements is not None:
            for val in prepared_attachements.values():
                if not hasattr(val, "__iter__"):
                    val = [val]
                for ob in val:
                    if isinstance(ob, BaseAttached):
                        s += ob.get_size()
        else:
            for attached in self.attached_names:
                for ob in getattr(self.associated, attached).all():
                    s += ob.get_size()
        return s

    def get_priority(self):
        return 0

    @classmethod
    def feature_urls(cls, name):
        """ For implementing component features """
        return []

    def get_content_name(self):
        return "{}_{}".format(
            self.localize_name(self.associated.ctype.name),
            self.associated_id
        )

    def get_content_description(self):
        return " "

    def localized_description(self):
        """ localize and perform other transforms before rendering to user """
        if not self.expose_description:
            return self.associated.description
        return gettext(self.associated.description)

    def get_strength(self):
        """ get required strength """
        return self.associated.ctype.strength

    def get_strength_link(self):
        """ get required strength for links """
        return self.get_strength()

    def get_abilities(self, context):
        """ Override for declaring content extra abilities """
        return set()

    def get_propagate_modified(self):
        return "\x1eunlisted\x1e" not in self.associated.info

    def get_form_kwargs(
        self, request, instance=None, disable_data=False, **kwargs
    ):
        """Return the keyword arguments for instantiating the form."""
        fkwargs = {}
        if instance:
            fkwargs["instance"] = instance
        else:
            fkwargs["instance"] = self

        if not disable_data and request.method in ('POST', 'PUT'):
            fkwargs.update({
                'data': request.POST,
                'files': request.FILES,
            })
        return fkwargs

    # default scopes:
    #  public accessible:
    #   raw: view with raw parameter
    #   list: raw but embedded in content list
    #   view: view of conten without raw parameter
    #  only accessible by owner and staff:
    #   add: create view of content
    #   update: update view of content
    #   raw_update: special update mode for machines or really fancy updates
    #   export: like raw just optimized for export: e.g. no dereferencing
    def get_form(self, scope):
        raise NotImplementedError

    def get_template_name(self, scope):
        if scope in ["add", "update"]:
            return 'spider_base/partials/base_form.html'
        return 'spider_base/view_form.html'

    def update_used_space(self, size_diff):
        if size_diff == 0:
            return
        f = "local"
        if VariantType.component_feature in self.associated.ctype.ctype:
            f = "remote"
        with transaction.atomic():
            self.associated.usercomponent.user_info.update_with_quota(
                    size_diff, f
                )
            self.associated.usercomponent.user_info.save(
                update_fields=[
                    "used_space_local", "used_space_remote"
                ]
            )

    def render_form(self, scope, **kwargs):
        _ = gettext
        if scope == "add":
            kwargs["form_empty_message"] = _("<b>No User Input required</b>")
            old_size = 0
        else:
            old_size = self.get_size()
        parent_form = kwargs.get("form", None)
        kwargs["form"] = self.get_form(scope)(
            **self.get_form_kwargs(
                scope=scope,
                **kwargs
            )
        )
        if kwargs["form"].is_valid():
            instance = kwargs["form"].save(False)
            try:
                self.update_used_space(
                    instance.get_size(self.prepared_attachements) - old_size
                )
            except ValidationError as exc:
                kwargs["form"].add_error(None, exc)
                messages.error(
                    kwargs["request"], _('Space exhausted')
                )
        if kwargs["form"].is_valid():
            instance.save()
            kwargs["form"].save_m2m()
            messages.success(
                kwargs["request"], _('Content updated')
            )
            kwargs["form"] = self.get_form(scope)(
                **self.get_form_kwargs(
                    scope=scope,
                    instance=instance,
                    disable_data=True,  # required for using object data
                    **kwargs
                )
            )

        elif parent_form:
            if len(kwargs["form"].errors.get(NON_FIELD_ERRORS, [])) > 0:
                parent_form.add_error(
                    None, kwargs["form"].errors.pop(NON_FIELD_ERRORS)
                )
            if self.associated_errors:
                parent_form.add_error(
                    None, self.associated_errors
                )
        return {
            "content": render_to_string(
                self.get_template_name(scope),
                request=kwargs["request"], context=kwargs
            ),
            "media": kwargs["form"].media
        }

    def get_absolute_url(self, scope="view"):
        return self.associated.get_absolute_url(scope)

    def get_primary_anchor(self, graph, context):
        p = self.associated.usercomponent.primary_anchor
        if p:
            return urljoin(
                get_anchor_domain(),
                p.get_absolute_url()
            )

        if not p and self.associated.usercomponent.public:
            return urljoin(
                get_anchor_domain(),
                self.associated.usercomponent.get_absolute_url()
            )
        return urljoin(
            get_anchor_domain(), context["request"].path
        )

    def get_references(self):
        return []

    def map_data(self, name, field, data, graph, context):
        if isinstance(data, File):
            return get_settings_func(
                "SPIDER_FILE_EMBED_FUNC",
                "spkcspider.apps.spider.functions.embed_file_default"
            )(name, data, self, context)
        return literalize(data, field, domain_base=context["hostpart"])

    def serialize(self, graph, ref_content, context):
        # context may not be updated here
        form = self.get_form(context["scope"])(
            **self.get_form_kwargs(
                disable_data=True,
                **context
            )
        )

        # context["abilities"] cache available if accessed via access
        # cannot update without breaking serializing, embed
        if "abilities" in context:
            abilities = context["abilities"]
        else:
            abilities = set(self.get_abilities(context))

        for ability in abilities:
            assert(ability not in default_abilities)

            graph.add((
                ref_content,
                spkcgraph["ability:name"],
                Literal(ability, datatype=XSD.string)
            ))

        for name, field in form.fields.items():
            raw_value = form.initial.get(name, None)
            try:
                value = field.to_python(raw_value)
            except Exception as exc:
                # user can corrupt tags, so just debug
                level = logging.WARNING
                if getattr(form, "layout_generating_form", False):
                    level = logging.DEBUG
                logger.log(
                    level,
                    "Corrupted field: %s, form: %s, error: %s",
                    name, form, exc
                )
                continue
            value_node = BNode()
            hashable = getattr(field, "hashable", False)

            graph.add((
                ref_content,
                spkcgraph["properties"],
                value_node
            ))
            graph.add((
                value_node,
                spkcgraph["hashable"],
                Literal(hashable)
            ))
            graph.add((
                value_node,
                spkcgraph["name"],
                Literal(name, datatype=XSD.string)
            ))
            graph.add((
                value_node,
                spkcgraph["fieldname"],
                Literal(form.add_prefix(name), datatype=XSD.string)
            ))

            if not isinstance(value, (list, tuple, models.QuerySet)):
                value = [value]

            for i in value:
                graph.add((
                    value_node,
                    spkcgraph["value"],
                    self.map_data(name, field, i, graph, context)
                ))
            if not value:
                graph.add((
                    value_node,
                    spkcgraph["value"],
                    RDF.nil
                ))

    def render_serialize(self, **kwargs):
        from ..models import AssignedContent
        # ** creates copy of dict, so it is safe to overwrite kwargs here

        session_dict = {
            "request": kwargs["request"],
            "context": kwargs,
            "scope": kwargs["scope"],
            "hostpart": kwargs["hostpart"],
            "ac_namespace": spkcgraph["contents"],
            "sourceref": URIRef(urljoin(
                kwargs["hostpart"], kwargs["request"].path
            ))
        }

        g = Graph()
        g.namespace_manager.bind("spkc", spkcgraph, replace=True)

        p = paginate_stream(
            AssignedContent.objects.filter(id=self.associated_id),
            getattr(
                settings, "SPIDER_SERIALIZED_PER_PAGE",
                settings.SPIDER_OBJECTS_PER_PAGE
            ),
            settings.SPIDER_MAX_EMBED_DEPTH
        )
        page = 1
        try:
            page = int(session_dict["request"].GET.get("page", "1"))
        except Exception:
            pass
        serialize_stream(
            g, p, session_dict,
            page=page,
            embed=True
        )
        if hasattr(kwargs["request"], "token_expires"):
            session_dict["expires"] = kwargs["request"].token_expires.strftime(
                "%a, %d %b %Y %H:%M:%S %z"
            )

        if page <= 1:
            source = kwargs.get("source", self)
            # "expires" (string) different from "token_expires" (datetime)
            if session_dict.get("expires"):
                add_property(
                    g, "token_expires", ob=session_dict["request"],
                    ref=session_dict["sourceref"]
                )
            if kwargs.get("machine_variants"):
                assert kwargs["request"].is_owner
                ucref = URIRef(urljoin(
                    kwargs["hostpart"],
                    source.associated.usercomponent.get_absolute_url()
                ))
                for machinec in kwargs["machine_variants"]:
                    g.add((
                        ucref,
                        spkcgraph["create:name"],
                        Literal(machinec, datatype=XSD.string)
                    ))
            g.add((
                session_dict["sourceref"],
                spkcgraph["scope"],
                Literal(kwargs["scope"], datatype=XSD.string)
            ))

            uc = kwargs.get("source", self.associated.usercomponent)
            g.add((
                session_dict["sourceref"], spkcgraph["strength"],
                Literal(uc.strength, datatype=XSD.integer)
            ))
            token = getattr(session_dict["request"], "auth_token", None)
            if token:
                token = token.token
            url2 = merge_get_url(str(session_dict["sourceref"]), token=token)
            g.add(
                (
                    session_dict["sourceref"],
                    spkcgraph["action:view"],
                    URIRef(url2)
                )
            )
            if kwargs["referrer"]:
                g.add((
                    session_dict["sourceref"],
                    spkcgraph["referrer"],
                    Literal(kwargs["referrer"], datatype=XSD.anyURI)
                ))
            if kwargs["token_strength"]:
                add_property(
                    g, "token_strength", ref=session_dict["sourceref"],
                    literal=kwargs["token_strength"]
                )
            add_property(
                g, "intentions", ref=session_dict["sourceref"],
                literal=kwargs["intentions"], datatype=XSD.string,
                iterate=True
            )

        ret = HttpResponse(
            g.serialize(format="turtle"),
            content_type="text/turtle;charset=utf-8"
        )

        if session_dict.get("expires", None):
            ret['X-Token-Expires'] = session_dict["expires"]
        # allow cors requests for raw
        ret["Access-Control-Allow-Origin"] = "*"
        return ret

    def access_view(self, **kwargs):
        _ = gettext

        kwargs["form"] = self.get_form("view")(
            **self.get_form_kwargs(disable_data=True, **kwargs)
        )
        kwargs.setdefault("legend", escape(_("View")))
        # not required: done by access template
        return {
            "content": render_to_string(
                self.get_template_name(kwargs["scope"]),
                request=kwargs["request"],
                context=kwargs
            ),
            "media": kwargs["form"].media
        }

    def access_add(self, **kwargs):
        _ = gettext
        kwargs.setdefault(
            "legend",
            escape(_("Add \"%s\"") % self.__str__())
        )
        # not visible by default
        kwargs.setdefault("confirm", _("Create"))
        # prevent second button
        kwargs.setdefault("inner_form", True)
        return self.render_form(**kwargs)

    def access_update(self, **kwargs):
        _ = gettext
        kwargs.setdefault(
            "legend",
            escape(_("Update \"%s\"") % self.__str__())
        )
        # not visible by default
        kwargs.setdefault("confirm", _("Update"))
        # prevent second button
        kwargs.setdefault("inner_form", True)
        return self.render_form(**kwargs)

    def access_export(self, **kwargs):
        kwargs["scope"] = "export"
        if VariantType.no_export in self.associated.ctype.ctype:
            raise Http404()
        return self.render_serialize(**kwargs)

    def access_raw(self, **kwargs):
        kwargs["scope"] = "raw"
        return self.render_serialize(**kwargs)

    def access_anchor(self, **kwargs):
        """
            Called for anchors
            WARNING: this means the context is completely different
                     "access", "get_abilities" method of anchors
                     (and links on them) must be robust
            WARNING: it is expected that get_abilities returns "anchor"
            WARNING: it is expected that access_anchor returns a HttpResponse
        """
        raise NotImplementedError()

    @csrf_exempt
    def access_default(self, **kwargs):
        raise Http404()

    access_raw_update = access_default

    def access(self, context):
        # context is updated and used outside!!
        # so make sure that func gets only a copy (**)
        context["abilities"] = set(self.get_abilities(context))
        context.setdefault("extra_outer_forms", [])
        func = self.access_default
        if context["scope"] == "view" and "raw" in context["request"].GET:
            func = self.access_raw
        elif context["scope"] in default_abilities:
            func = getattr(self, "access_{}".format(context["scope"]))
        elif context["scope"] in context["abilities"]:
            func = getattr(self, "access_{}".format(context["scope"]))

        # check csrf tokens manually
        if not getattr(func, "csrf_exempt", False):
            csrferror = CsrfViewMiddleware().process_view(
                context["request"], None, (), {}
            )
            if csrferror is not None:
                # csrferror is HttpResponse
                context["content"] = csrferror
                return context
        ret = func(**context)
        # update function should never return HttpResponse for GET
        if isinstance(ret, HttpResponseBase):
            assert (
                context["scope"] != "update" or
                context["request"].method != "GET"
            )
            context["content"] = ret
        elif isinstance(ret, str):
            context["content"] = ret
        elif isinstance(ret, dict):
            oldmedia = context["media"]
            context.update(ret)
            context["media"] += oldmedia
        else:
            raise NotImplementedError()
        return context

    def get_info(self, unique=None, unlisted=None):
        # unique=None, feature=None shortcuts for get_info overwrites
        # passing down these parameters not neccessary
        if unique is None:
            unique = (
                VariantType.unique in self.associated.ctype.ctype
            )
        if unlisted is None:
            unlisted = (
                VariantType.component_feature in
                self.associated.ctype.ctype or
                VariantType.content_feature in
                self.associated.ctype.ctype
            )

        anchortag = ""
        if VariantType.anchor in self.associated.ctype.ctype:
            anchortag = "anchor\x1e"

        idtag = "primary\x1e"
        # simulates beeing not unique, by adding id
        if not unique:
            idtag = "id=\x1e"  # placeholder
            if getattr(self.associated, "id", None):
                idtag = "id={}\x1e".format(self.associated_id)
        unlistedtag = ""
        if unlisted:
            unlistedtag = "unlisted\x1e"
        return "\x1etype={}\x1e{}{}{}".format(
            self.associated.ctype.name,
            idtag,
            anchortag,
            unlistedtag
        )

    def full_clean(self, **kwargs):
        # checked with clean
        kwargs.setdefault("exclude", []).append("associated")
        return super().full_clean(**kwargs)

    def update_associated(self):
        if getattr(self, "id", None):
            self.associated.info = self.get_info()
            if not self.expose_name or not self.associated.name:
                self.associated.name = self.get_content_name()
            if not self.expose_description or not self.associated.description:
                self.associated.description = self.get_content_description()
        self.associated.priority = self.get_priority()
        self.associated.strength = self.get_strength()
        self.associated.strength_link = self.get_strength_link()

    def clean(self):
        _ = gettext
        self.update_associated()
        try:
            self.associated.full_clean(exclude=["datacontent"])
        except ValidationError as exc:
            self.associated_errors = exc
        if self.associated_errors:
            raise ValidationError(
                _('AssignedContent validation failed'),
                code="assigned_content"
            )
        self._content_is_cleaned = True

    def save(self, *args, **kwargs):
        if settings.DEBUG:
            assert self._content_is_cleaned, "try to save uncleaned content"
        created = False
        assignedcontent = self.associated
        if not getattr(self, "id", None):
            created = True
            # create the model for saving BaseContent
            assignedcontent.save()
        # this changes the associated model for no reason
        super().save(*args, **kwargs)
        # restore it here
        self.associated = assignedcontent
        if created:
            # add requires this
            self.update_associated()
        if not self.associated.token:
            if self.force_token_size:
                self.associated.token_generate_new_size = \
                    self.force_token_size
            else:
                self.associated.token_generate_new_size = \
                    getattr(settings, "INITIAL_STATIC_TOKEN_SIZE", 30)
        # set token
        if self.associated.token_generate_new_size is not None:
            if self.associated.token:
                print(
                    "Old nonce for Content id:", self.associated_id,
                    "is", self.associated.token
                )
            self.associated.token = create_b64_id_token(
                self.associated_id,
                "_",
                self.associated.token_generate_new_size
            )
            self.associated.token_generate_new_size = None
        # save associated
        self.associated.save()

        if self.prepared_attachements:
            for key, val in self.prepared_attachements.items():
                if not hasattr(val, "__iter__"):
                    val = [val]
                pks = set()
                for i in val:
                    i.save()
                    pks.add(i.pk)
                # remove rest (if foreignkey)
                fieldmanager = getattr(self.associated, key)
                field = self.associated._meta.get_field(key)
                if field.many_to_many:
                    fieldmanager.set(val)
                elif field.one_to_many:
                    fieldmanager.exclude(pk__in=pks).delete()
        # needs id first
        s = set(self.associated.attached_contents.all())
        s.update(self.get_references())
        self.associated.references.set(s)
        # message usercomponent about change
        if self.get_propagate_modified():
            self.associated.usercomponent.save(update_fields=["modified"])
        # delete saved errors
        self.associated_errors = None
        # require cleaning again
        self._content_is_cleaned = False
        # reset to default
        self.prepared_attachements = None



# Generated by Django 2.2.5 on 2019-09-29 21:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_verifier', '0002_20190530_squashed'),
    ]

    operations = [
        migrations.AlterField(
            model_name='verifysourceobject',
            name='token',
            field=models.CharField(blank=True, max_length=255, null=True, unique=True),
        ),
    ]



# Generated by Django 3.0.1 on 2019-12-30 13:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0012_auto_20191230_1305'),
        ('spider_webcfg', '0004_auto_20191230_1351'),
    ]

    operations = [
        migrations.DeleteModel(
            name='WebConfig',
        ),
        migrations.CreateModel(
            name='WebConfig',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('spider_base.datacontent',),
        ),
    ]



__all__ = [
    "DataContent", "BaseAttached", "AttachedFile", "AttachedTimespan",
    "AttachedBlob"
]

import posixpath

from django.conf import settings
from django.core.files.storage import default_storage
from django.db import models
from django.core.exceptions import ValidationError
from django.utils.translation import gettext
from django.http import HttpResponseRedirect
from jsonfield import JSONField
from ranged_response import RangedFileResponse

from spkcspider.utils.security import create_b64_token

from ..abstract_models import BaseContent
from ..conf import FILE_TOKEN_SIZE


def get_file_path(instance, filename):
    ret = getattr(settings, "SPIDER_FILE_DIR", "spider_files")
    # try 100 times to find free filename
    # but should not take more than 1 try
    # IMPORTANT: strip . to prevent creation of htaccess files or similar
    for _i in range(0, 100):
        ret_path = default_storage.generate_filename(
            posixpath.join(
                ret, str(instance.content.usercomponent.user_id),
                create_b64_token(FILE_TOKEN_SIZE), filename.lstrip(".")
            )
        )
        if not default_storage.exists(ret_path):
            break
    else:
        raise FileExistsError("Unlikely event: no free filename")
    return ret_path


class DataContent(BaseContent):
    """
        inherit from it, it will maybe replace the generic relation
    """
    # has as an exception a related name (e.g. for speeding up)
    associated = models.OneToOneField(
        "spider_base.AssignedContent",
        on_delete=models.CASCADE, null=True
    )
    quota_data = JSONField(default=dict, blank=True)
    free_data = JSONField(default=dict, blank=True)

    def get_size(self, prepared_attachements=None):
        s = super().get_size(prepared_attachements)
        s += len(str(self.quota_data))
        return s


class BaseAttached(models.Model):
    id: int = models.BigAutoField(primary_key=True, editable=False)
    name = models.CharField(max_length=50, default="", blank=True)
    unique = models.BooleanField(default=False, blank=True)
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)
    content = models.ForeignKey(
        "spider_base.AssignedContent",
        on_delete=models.CASCADE,
        related_name="%(class)ss"
    )

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not getattr(cls._meta, "abstract", False):
            BaseContent.attached_names.add(
                "{}_set".format(cls._meta.model_name)
            )

    class Meta:
        abstract = True
        constraints = [
            models.UniqueConstraint(
                name="%(class)s_unique",
                fields=("content", "name"),
                condition=models.Q(unique=True)
            )
        ]


class AttachedTimespan(BaseAttached):
    start = models.DateTimeField()
    stop = models.DateTimeField(blank=True, null=True)

    class Meta:
        constraints = [
            models.CheckConstraint(
                check=models.Q(
                    start__lte=models.F("stop")
                ) |
                models.Q(start__isnull=True) |
                models.Q(stop__isnull=True),
                name="%(class)s_order"
            ),
            models.CheckConstraint(
                check=models.Q(
                    start__isnull=False
                ) |
                models.Q(
                    stop__isnull=False
                ),
                name="%(class)s_exist"
            )
        ] + BaseAttached.Meta.constraints

    def clean(self):
        _ = gettext
        if self.start and self.stop:
            if self.stop < self.start:
                raise ValidationError(
                    _("Stop (%(stop)s) < Start (%(start)s)"),
                    params={
                        "start": self.start,
                        "stop": self.stop
                    }
                )
        elif not self.start and not self.stop:
            raise ValidationError(
                _("Timespan empty"),
            )

    def get_size(self):
        # free can be abused
        return 8


class AttachedCounter(BaseAttached):
    counter = models.BigIntegerField()

    def get_size(self):
        # free can be abused
        return 4


class AttachedFile(BaseAttached):
    file = models.FileField(upload_to=get_file_path, null=False, blank=False)

    def get_response(self, request=None, name=None, add_extension=False):
        if not request:
            response = HttpResponseRedirect(
                redirect_to=self.file.url
            )
        else:
            response = RangedFileResponse(
                request,
                self.file.file,
                content_type='application/octet-stream'
            )
            if not name:
                name = posixpath.basename(self.file.name)
            if add_extension and "." not in name:  # use ending of saved file
                ext = self.file.name.rsplit(".", 1)
                if len(ext) > 1:
                    name = "%s.%s" % (name, ext[1])
            # name is sanitized to not contain \n, and other ugly control chars
            response['Content-Disposition'] = \
                'attachment; filename="%s"' % posixpath.basename(name.replace(
                    r'"', r'\"'
                ))
        return response

    def get_size(self):
        return self.file.size

    def save(self, *args, **kw):
        if self.pk is not None:
            orig = AttachedFile.objects.get(pk=self.pk)
            if orig.file != self.file:
                orig.file.delete(False)
        super().save(*args, **kw)


class AttachedBlob(BaseAttached):
    blob = models.BinaryField(default=b"", editable=True, blank=True)

    def get_size(self):
        return len(self.blob)

    @property
    def as_bytes(self):
        return bytes(self.blob)



__all__ = [
    "Registry", "ProtectionRegistry", "ContentRegistry", "FeatureUrlsRegistry",
    "ContentDeletionPeriodRegistry",
    "protections", "contents", "feature_urls", "content_deletion_periods"
]

import logging

from django.shortcuts import resolve_url

from spkcspider.constants import ActionUrl

logger = logging.getLogger(__name__)


class Registry(object):
    key_type_registry = {
        "str": lambda x: x
    }
    attr_path = None

    _unchecked_apps = None
    _populated = False
    _registry = None

    find_func = None

    def __init__(self):
        self._unchecked_apps = set()
        self._registry = {}

    def __getitem__(self, ob):
        key = self.get_key(ob)
        if key not in self._registry:
            find_func = self.find_func or self.generate_find_func
            self._registry[key] = find_func(key)
        return self._registry[key]

    def __setitem__(self, key, value):
        self._registry[key] = value

    def __contains__(self, ob):
        return self.get_key(ob) in self._registry

    def __getattr__(self, attr):
        return getattr(self._registry, attr)

    def get_key(self, ob):
        return self.key_type_registry[type(ob).__name__](ob)

    def generate_find_func(self, key):
        """ generate generator based find function """
        from django.apps import apps
        from importlib import import_module
        self._unchecked_apps.update(apps.get_app_configs())

        def find_func(key):
            if self._unchecked_apps:
                checked = []
                for app in self._unchecked_apps:
                    if hasattr(app, self.attr_path):
                        import_module(
                            getattr(app, self.attr_path),
                            app.name
                        )
                    checked.append(app)
                    if key in self:
                        break
                self._unchecked_apps.difference_update(checked)
            return self[key]
        self.find_func = find_func
        return self.find_func(key)

    def populate(self):
        """ Populate all possible content, required for iterations """
        from importlib import import_module
        from django.apps import apps
        for app in apps.get_app_configs():
            if hasattr(app, self.attr_path):
                import_module(
                    getattr(app, self.attr_path),
                    app.name
                )
        self._populated = True

    def keys(self):
        if not self._populated:
            self.populate()
        return self._registry.keys()

    def values(self):
        if not self._populated:
            self.populate()
        return self._registry.values()

    def items(self):
        if not self._populated:
            self.populate()
        return self._registry.items()


class ProtectionRegistry(Registry):
    _unchecked_apps = None
    attr_path = "spider_protections_path"

    def __init__(self):
        self.key_type_registry = {
            "str": lambda x: x,
            "Protection": lambda x: x.code,
            "AssignedProtection": lambda x: x.protection.code
        }
        super().__init__()

    def __setitem__(self, key, value):
        if key in self:
            raise Exception("Already exists")
        assert value, "invalid value type %s" % type(value)
        super().__setitem__(key, value)

    def initialize(self):
        from spkcspider.constants import ProtectionStateType
        from django.apps import apps
        UserComponent = apps.get_model("spider_base", "UserComponent")
        AssignedProtection = apps.get_model(
            "spider_base",
            "AssignedProtection"
        )
        ProtectionModel = apps.get_model("spider_base", "Protection")
        # auto initializes
        for code, val in self.items():
            ret = ProtectionModel.objects.get_or_create(
                defaults={"ptype": val.ptype}, code=code
            )[0]
            if ret.ptype != val.ptype:
                ret.ptype = val.ptype
                ret.save()

        login = ProtectionModel.objects.filter(code="login").first()
        if login:
            for uc in UserComponent.objects.filter(name="index"):
                assignedprotection = AssignedProtection.objects.get_or_create(
                    defaults={"state": ProtectionStateType.enabled},
                    usercomponent=uc, protection=login
                )[0]
                if assignedprotection.state == ProtectionStateType.disabled:
                    assignedprotection.state = ProtectionStateType.enabled
                    assignedprotection.save()

        invalid_models = ProtectionModel.objects.exclude(
            code__in=self.keys()
        )
        if invalid_models.exists():
            print(
                "Invalid protections, please update or remove them:",
                [t.code for t in invalid_models]
            )


class ContentRegistry(Registry):
    # updated attributes of ContentVariant
    _attribute_list = ["name", "ctype", "strength"]

    def __init__(self):
        self.key_type_registry = {
            "str": lambda x: x,
            "ContentVariant": lambda x: x.code,
            "AssignedContent": lambda x: x.ctype.code
        }
        super().__init__()

    def __setitem__(self, key, value):
        if key in self:
            raise Exception("Already exists")
        assert value, "invalid value type %s" % type(value)
        super().__setitem__(key, value)

    def generate_find_func(self, key):
        if not self._populated:
            self.populate()

        def find_func(key):
            return self[key]
        self.find_func = find_func
        return self.find_func(key)

    def populate(self):
        from django.apps import apps
        apps.populate()
        self._populated = True

    def initialize(self):
        from spkcspider.constants import essential_contents, VariantType
        from django.db import models, transaction
        from django.db.utils import IntegrityError
        from django.apps import apps
        from .abstract_models.contents import forbidden_names
        ContentVariant = apps.get_model("spider_base", "ContentVariant")

        all_content = models.Q()
        valid_for = {}

        diff = essential_contents.difference(
            self.keys()
        )
        if diff:
            logger.warning(
                "Missing essential contents: %s" % diff
            )

        for code, val in self.items():
            appearances = val.appearances
            if callable(appearances):
                appearances = appearances()
            assert appearances is not None, "missing appearances: %s" % code

            # update name if only one name exists
            update = (len(appearances) == 1)

            for attr_dict in appearances:
                require_save = False
                assert attr_dict["name"] not in forbidden_names, \
                    "Forbidden content name: %s" % attr_dict["name"]
                attr_dict = attr_dict.copy()
                _v_for = attr_dict.pop("valid_feature_for", None)
                try:
                    with transaction.atomic():
                        if update:
                            variant = ContentVariant.objects.get_or_create(
                                defaults=attr_dict, code=code
                            )[0]
                        else:
                            variant = ContentVariant.objects.get_or_create(
                                defaults=attr_dict, code=code,
                                name=attr_dict["name"]
                            )[0]

                except IntegrityError:
                    # renamed model = code changed
                    variant = ContentVariant.objects.get(
                        name=attr_dict["name"]
                    )
                    variant.code = code
                    require_save = True

                if _v_for:
                    if _v_for == "*":
                        valid_for[attr_dict["name"]] = (variant, set("*"))
                    else:
                        valid_for[attr_dict["name"]] = (variant, set(_v_for))
                elif VariantType.content_feature in variant.ctype:
                    logger.warning(
                        "%s defines content_feature but defines no "
                        "\"valid_feature_for\"", variant.name
                    )

                for key in self._attribute_list:
                    val = attr_dict.get(
                        key, variant._meta.get_field(key).get_default()
                    )
                    if getattr(variant, key) != val:
                        setattr(variant, key, val)
                        require_save = True
                if require_save:
                    variant.save()
                all_content |= models.Q(name=attr_dict["name"], code=code)
        for val in valid_for.values():
            try:
                # try first to exclude by stripping "*" and checking error
                val[1].remove("*")
                val[0].valid_feature_for.set(ContentVariant.objects.exclude(
                    name__in=val[1]
                ))
            except KeyError:
                val[0].valid_feature_for.set(ContentVariant.objects.filter(
                    name__in=val[1]
                ))

        invalid_models = ContentVariant.objects.exclude(all_content)
        if invalid_models.exists():
            print(
                "Invalid content, please update or remove them:",
                ["\"{}\":{}".format(t.code, t.name) for t in invalid_models]
            )


class FeatureUrlsRegistry(Registry):
    contentRegistry = None
    default_actions = None

    def __init__(self, contentRegistry):
        self.contentRegistry = contentRegistry
        self.default_actions = {}
        self.key_type_registry = {
            "tuple": lambda x: x,
            "list": tuple,
            "ContentVariant": lambda x: (x.code, x.name),
            "AssignedContent": lambda x: (x.ctype.code, x.ctype.name)
        }
        super().__init__()

    def generate_find_func(self, key):
        from django.conf import settings
        self.default_actions["delete-token"] = \
            "spider_base:token-delete-request"
        if getattr(settings, "DOMAINAUTH_URL", None):
            self.default_actions["domainauth-url"] = settings.DOMAINAUTH_URL

        def find_func(key):
            self[key] = frozenset(map(
                lambda x: ActionUrl(x[0], resolve_url(x[1])),
                self.contentRegistry[key[0]].feature_urls(key[1])
            ))
            return self[key]
        self.find_func = find_func
        return self.find_func(key)

    def populate(self):
        for code, content in self.contentRegistry.items():
            for appearance in content.appearances:
                # calls generate_find_func if required
                self[(code, appearance["name"])]
        self._populated = True


class ContentDeletionPeriodRegistry(Registry):
    contentRegistry = None

    def __init__(self, contentRegistry):
        self.contentRegistry = contentRegistry
        self.key_type_registry = {
            "tuple": lambda x: x,
            "ContentVariant": lambda x: (x.code, x.name),
            "AssignedContent": lambda x: (x.ctype.code, x.ctype.name)
        }
        super().__init__()

    def generate_find_func(self, key):
        from django.conf import settings
        self.update(
            getattr(
                settings, "SPIDER_CONTENTS_DELETION_PERIODS", {}
            )
        )

        def find_func(key):
            deletion_period = self.contentRegistry[key[0]].deletion_period
            if callable(deletion_period):
                self[key] = deletion_period(key[1])
            else:
                self[key] = deletion_period
            return self[key]
        self.find_func = find_func
        return self.find_func(key)

    def populate(self):
        for code, content in self.contentRegistry.items():
            for appearance in content.appearances:
                # calls generate_find_func if required
                self[(code, appearance["name"])]
        self._populated = True


protections = ProtectionRegistry()
contents = ContentRegistry()
feature_urls = FeatureUrlsRegistry(contents)
content_deletion_periods = ContentDeletionPeriodRegistry(contents)



"""
    RDF namespaces and related constants
"""

__all__ = [
    "spkcgraph",
]

from rdflib.namespace import Namespace

# Literal allows arbitary datatypes, use this and don't bind
spkcgraph = Namespace("https://spkcspider.net/static/schemes/spkcgraph#")



# Generated by Django 3.0.2 on 2020-01-04 17:36

from django.db import migrations


def move_to_datacontent(apps, schema_editor):
    AssignedContent = apps.get_model("spider_base", "AssignedContent")
    UserTagLayout = apps.get_model("spider_tags", "UserTagLayout")
    TagLayout = apps.get_model("spider_tags", "TagLayout")
    SpiderTag = apps.get_model("spider_tags", "SpiderTag")
    DataContent = apps.get_model("spider_base", "DataContent")

    for a in AssignedContent.objects.filter(
        ctype__code=UserTagLayout._meta.model_name
    ):
        d = DataContent(associated=a)
        layout = TagLayout.objects.get(usertag__id=a.object_id)
        layout.usertag = None
        layout.save()
        d.free_data["tmp_layout_id"] = layout.id
        d.save()

    for a in AssignedContent.objects.filter(
        ctype__code=SpiderTag._meta.model_name
    ):
        content = SpiderTag.objects.get(id=a.object_id)
        content.associated = a
        content.save()


class Migration(migrations.Migration):

    dependencies = [
        ('spider_tags', '0008_auto_20200104_1735'),
    ]

    run_before = [
        ('spider_base', '0012_auto_20191230_1305'),
    ]

    operations = [
        migrations.RunPython(move_to_datacontent),
    ]



__all__ = ["WebConfigForm"]

from django import forms
from spkcspider.apps.spider.forms.base import DataContentForm


class WebConfigForm(DataContentForm):
    creation_url = forms.URLField(disabled=True, required=False)
    config = forms.CharField(
        widget=forms.Textarea()
    )

    free_fields = {"creation_url": None}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        b = None
        if self.object.id:
            b = self.object.associated.attachedblobs.filter(
                name="config"
            ).first()
        if b:
            self.initial["config"] = b.blob.as_bytes.decode("ascii")



"""
User Components/Info
namespace: spider_base

"""

__all__ = [
    "UserComponent", "UserComponentManager", "TokenCreationError", "UserInfo"
]

import logging
import datetime

from django.apps import apps
from django.conf import settings
from django.core import validators
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from spkcspider.constants import (
    MAX_TOKEN_B64_SIZE, ProtectionType, TokenCreationError, VariantType,
    hex_size_of_bigid, static_token_matcher
)
from spkcspider.utils.security import create_b64_id_token
from spkcspider.utils.settings import get_settings_func

from ..conf import default_uctoken_duration, force_captcha
from ..validators import validator_token
from .. import registry

logger = logging.getLogger(__name__)


_name_help = _(
    "Name of the component.<br/>"
    "Note: there are special named components "
    "with different protection types and scopes.<br/>"
    "Most prominent: \"index\" for authentication"
)


_required_passes_help = _(
    "How many protection must be passed? "
    "Set greater 0 to enable protection based access"
)

_feature_help = _(
    "Appears as featured on \"home\" page"
)

_0td = datetime.timedelta()


class UserComponentManager(models.Manager):
    def _update_args(self, defaults, kwargs):
        if defaults is None:
            defaults = {}
        if kwargs is None:
            kwargs = defaults
        name = kwargs.get("name", defaults.get("name", None))
        if name == "index" and force_captcha:
            defaults["required_passes"] = 2
            defaults["strength"] = 10
        elif name == "index":
            defaults["required_passes"] = 1
            defaults["strength"] = 10
        elif kwargs.get("public", defaults.get("public", False)):
            if kwargs.get(
                "required_passes",
                defaults.get("required_passes", 0)
            ) == 0:
                defaults["strength"] = 0
            else:
                defaults["strength"] = 4
        else:
            if kwargs.get(
                "required_passes",
                defaults.get("required_passes", 0)
            ) == 0:
                defaults["strength"] = 5
            else:
                defaults["strength"] = 9
        return defaults

    def create(self, **kwargs):
        ret = self.get_queryset().create(
            **self._update_args(kwargs, None)
        )
        if not ret.token:
            ret.token = create_b64_id_token(ret.id, "_")
            ret.save(update_fields=["token"])
        return ret

    def update_or_create(self, defaults=None, **kwargs):
        ret = self.get_queryset().update_or_create(
            defaults=self._update_args(defaults, kwargs), **kwargs
        )
        if not ret[0].token:
            ret[0].token = create_b64_id_token(ret[0].id, "_")
            ret[0].save(update_fields=["token"])
        return ret

    def get_or_create(self, defaults=None, **kwargs):
        ret = self.get_queryset().get_or_create(
            defaults=self._update_args(defaults, kwargs), **kwargs
        )
        if not ret[0].token:
            ret[0].token = create_b64_id_token(ret[0].id, "_")
            ret[0].save(update_fields=["token"])
        return ret

    def from_url_part(self, url):
        """ can be full url or token/accessmethod """
        res = static_token_matcher.match(url)
        if not res:
            raise self.model.DoesNotExist()
        res = res.groupdict()
        if res["access"] == "list":
            return self.get(
                token=res["static_token"]
            )
        else:
            return self.get(
                contents__token=res["static_token"]
            )


class UserComponent(models.Model):
    id: int = models.BigAutoField(primary_key=True, editable=False)
    # brute force protection and identifier, replaces nonce
    #  16 = usercomponent.id in hexadecimal
    #  +1 for seperator
    token: str = models.CharField(
        max_length=(MAX_TOKEN_B64_SIZE)+hex_size_of_bigid+2,
        db_index=True, unique=True, null=True, blank=True,
        validators=[
            validator_token
        ]
    )
    public: bool = models.BooleanField(
        default=False,
        help_text=_(
            "Is public? Is listed and searchable?<br/>"
            "Note: This field is maybe not deactivatable "
            "because of assigned content"
        )
    )
    # special name: index:
    #    protections are used for authentication
    #    attached content is only visible for admin and user
    # db_index=True: "index" requests can speed up
    # regex disables controlcars and disable special spaces
    name: str = models.CharField(
        max_length=255,
        null=False,
        db_index=True,
        help_text=_name_help,
        validators=[validators.RegexValidator(r"^(\w[\w ]*\w|\w?)$")]
    )
    description: str = models.TextField(
        default="",
        help_text=_(
            "Description of user component."
        ), blank=True
    )
    required_passes: int = models.PositiveIntegerField(
        default=0,
        help_text=_required_passes_help
    )
    # cached protection strength
    strength: int = models.PositiveSmallIntegerField(
        default=0,
        validators=[validators.MaxValueValidator(10)],
        editable=False
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE,
    )
    features = models.ManyToManyField(
        "spider_base.ContentVariant",
        related_name="feature_for_components", blank=True,
        limit_choices_to=models.Q(
            ctype__contains=VariantType.component_feature
        )
    )
    default_content_features = models.ManyToManyField(
        "spider_base.ContentVariant",
        related_name="default_feature_for_contents", blank=True,
        limit_choices_to=models.Q(
            ctype__contains=VariantType.content_feature
        ),
        help_text=_(
            "Select features used by default for contents in this "
            "component"
        )
    )
    primary_anchor = models.ForeignKey(
        "spider_base.AssignedContent", related_name="primary_anchor_for",
        null=True, blank=True,
        limit_choices_to={
            "info__contains": "\x1eanchor\x1e",
        }, on_delete=models.SET_NULL,
        help_text=_(
            "Select main identifying anchor. Also used for attaching "
            "persisting tokens (elsewise they are attached to component)"
        )
    )
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)

    # only admin
    featured: bool = models.BooleanField(
        default=False, help_text=_feature_help
    )

    # both should not be edited
    can_auth: bool = models.BooleanField(default=False, editable=False)

    token_duration = models.DurationField(
        default=default_uctoken_duration,
        null=False
    )
    # only editable for admins
    deletion_requested = models.DateTimeField(
        null=True, default=None, blank=True
    )
    # fix linter warning
    objects = UserComponentManager()
    contents = None
    # should be used for retrieving active protections, related_name
    protections = None

    class Meta:
        unique_together = [("user", "name")]
        permissions = [("can_feature", "Can feature User Components")]

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<UserComponent: (%s: %s)>" % (self.username, self.__str__())

    def get_component_quota(self):
        return get_settings_func(
            "SPIDER_GET_QUOTA",
            "spkcspider.apps.spider.functions.get_quota"
        )(self.user, "usercomponents")

    def clean(self):
        _ = gettext
        self.public = (self.public and self.is_public_allowed)
        self.featured = (self.featured and self.public)
        assert(self.is_index or self.strength < 10)

        obj = self.contents.filter(
            strength__gt=self.strength
        ).order_by("strength").last()
        if obj:
            raise ValidationError(
                _(
                    'Protection strength too low, required: %(strength)s'
                ),
                code="strength",
                params={'strength': obj.strength},
            )

    def auth(self, request, ptype=ProtectionType.access_control,
             protection_codes=None, side_effect=False,
             order_by="protection__code", **kwargs):
        # caching problems, will cause heisenbugs
        # if QuerySet is accessed via reverse foreign key
        # so use explicit import
        from .protections import AssignedProtection
        if side_effect:
            return AssignedProtection.objects.filter(
                usercomponent=self
            ).valid().order_by(
                order_by
            ).authall(
                request, required_passes=0,
                ptype=ptype, protection_codes=protection_codes,
                **kwargs
            )
        return AssignedProtection.objects.filter(
            usercomponent=self
        ).valid().order_by(
            order_by
        ).authall(
            request=request, required_passes=self.required_passes,
            ptype=ptype, protection_codes=protection_codes,
            **kwargs
        )

    def get_accumulated_size(self):
        _local_size = 0
        _remote_size = 0
        for elem in self.contents.all():
            if elem.ctype.is_feature:
                _remote_size += elem.get_size()
            else:
                _local_size += elem.get_size()
        return _local_size, _remote_size

    def get_absolute_url(self):
        return reverse(
            "spider_base:ucontent-list",
            kwargs={
                "token": self.token
            }
        )

    @property
    def untrusted_strength(self):
        if self.strength in {0, 4, 5, 9, 10}:
            return self.strength
        elif self.strength < 4:
            return 3
        elif self.strength < 9:
            return 8
        else:
            logger.warning("Invalid strength: %s", self.strength)

    @property
    def username(self):
        return getattr(self.user, self.user.USERNAME_FIELD)

    @property
    def user_info(self):
        return self.user.spider_info

    @property
    def index(self):
        return UserComponent.objects.get(user=self.user, name="index")

    @property
    def is_index(self) -> bool:
        return (self.name == "index")

    @property
    def is_public_allowed(self) -> bool:
        """ Is the public attribute allowed """
        return not self.is_index and not self.contents.filter(
            strength__gte=5
        ).exists()

    @property
    def deletion_period(self):
        return getattr(
            settings, "SPIDER_COMPONENTS_DELETION_PERIODS", {}
        ).get(self.name, None) or _0td


class UserInfo(models.Model):
    """ Contains generated Informations about user """
    id: int = models.BigAutoField(primary_key=True, editable=False)
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, editable=False,
        related_name="spider_info",
    )
    allowed_content = models.ManyToManyField(
        "spider_base.ContentVariant", related_name="+", editable=False
    )
    used_space_local: int = models.BigIntegerField(default=0, editable=False)
    used_space_remote: int = models.BigIntegerField(default=0, editable=False)

    class Meta:
        default_permissions = ()

    def calculate_allowed_content(self):
        ContentVariant = apps.get_model("spider_base.ContentVariant")
        allowed = []
        cfilterfunc = get_settings_func(
            "SPIDER_CONTENTVARIANT_FILTER",
            "spkcspider.apps.spider.functions.allow_all_filter"
        )
        # Content types which are not "installed" should be removed/never used
        # unlisted can be removed as sideproduct if not specified with feature
        # or machine
        for variant in ContentVariant.objects.exclude(
            ~(
                models.Q(ctype__contains=VariantType.content_feature) |
                models.Q(ctype__contains=VariantType.component_feature) |
                models.Q(ctype__contains=VariantType.machine)
            ),
            ctype__contains=VariantType.unlisted
        ).filter(
            code__in=registry.contents.keys()
        ):
            # always include special variants
            # elsewise unnecessary recalculations are done and other bugs
            if variant.name in {"DomainMode", "DefaultActions"}:
                allowed.append(variant)
            elif cfilterfunc(self.user, variant):
                allowed.append(variant)
        # save not required, m2m field
        self.allowed_content.set(allowed)

    def calculate_used_space(self):
        from . import AssignedContent
        self.used_space_local = 0
        self.used_space_remote = 0
        for c in AssignedContent.objects.filter(
            usercomponent__user=self.user
        ):
            if c.ctype.is_feature:
                self.used_space_remote += c.get_size()
            else:
                self.used_space_local += c.get_size()

    def update_with_quota(self, size_diff, quota_type):
        fname = "used_space_{}".format(quota_type)
        qval = getattr(self, fname)
        quota = get_settings_func(
            "SPIDER_GET_QUOTA",
            "spkcspider.apps.spider.functions.get_quota"
        )(self.user, quota_type)
        # if over quota: reducing size is always good and should never fail
        # be nice and also allow keeping the same amount of space
        if quota and size_diff > 0 and qval + size_diff > quota:
            raise ValidationError(
                _("Exceeds quota by %(diff)s Bytes"),
                code='quota_exceeded',
                params={'diff': size_diff},
            )
        setattr(
            self, fname, models.F(fname)+size_diff
        )




import logging
import posixpath

from django.conf import settings
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.html import escape
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext
from spkcspider.apps.spider.conf import (
    image_extensions, media_extensions
)
from spkcspider.apps.spider.models import DataContent
from spkcspider.apps.spider.abstract_models import BaseContent
from spkcspider.apps.spider import registry
from spkcspider.constants import VariantType
from spkcspider.utils.fields import prepare_description, add_by_field

from .conf import LICENSE_CHOICES

logger = logging.getLogger(__name__)


# Create your models here.


@add_by_field(registry.contents, "_meta.model_name")
class DisclaimerFeature(BaseContent):
    appearances = [
        {
            "name": "EUDisclaimer",
            "ctype": (
                VariantType.content_feature + VariantType.unlisted
            ),
            "valid_feature_for": ["Text"]
        }
    ]

    class Meta(BaseContent.Meta):
        abstract = True

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", name)


class LicenseMixin(object):
    license_name_translation_list = LICENSE_CHOICES

    @property
    def full_license_name(self):
        license_name = self.free_data.get("license_name", "other")
        return self.license_name_translation_list.get(
            license_name, {}
        ).get("name", license_name)


@add_by_field(registry.contents, "_meta.model_name")
class FileFilet(LicenseMixin, DataContent):
    expose_name = True
    expose_description = True

    appearances = [{"name": "File"}]

    class Meta:
        proxy = True

    def get_template_name(self, scope):
        if scope in ["add", "update"]:
            return 'spider_filets/file_form.html'
        return 'spider_base/view_form.html'

    def get_content_name(self):
        # in case no name is set

        if self.prepared_attachements:
            fob = self.prepared_attachements["attachedfiles"][0].file
        else:
            fob = self.associated.attachedfiles.get(name="file").file
        return posixpath.basename(fob.file.name)

    def get_info(self):
        ret = super().get_info()
        split = self.associated.name.rsplit(".", 1)
        ftype = "binary"
        if len(split) == 2:
            extension = split[1].lower()
            if extension in image_extensions:
                ftype = "image"
            elif extension in media_extensions:
                ftype = "media"
        return "%sname=%s\x1efile_type=%s\x1e" % (
            ret, self.associated.name, ftype
        )

    def get_form(self, scope):
        from .forms import FileForm
        return FileForm

    def get_abilities(self, context):
        return set(("download",))

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret["request"] = kwargs["request"]
        ret["uc"] = self.associated.usercomponent
        return ret

    def render_form(self, **kwargs):
        kwargs["enctype"] = "multipart/form-data"
        return super().render_form(**kwargs)

    def access_view(self, **kwargs):
        kwargs["object"] = self
        kwargs["file"] = self.associated.attachedfiles.get(name="file").file
        kwargs["type"] = self.associated.getlist("file_type", 1)
        if kwargs["type"]:
            kwargs["type"] = kwargs["type"][0]
        else:
            kwargs["type"] = None
        if getattr(settings, "FILE_DIRECT_DOWNLOAD", False):
            kwargs["download"] = kwargs["file"].url
        else:
            kwargs["download"] = "{}?{}".format(
                reverse(
                    'spider_base:ucontent-access',
                    kwargs={
                        "token": self.associated.token,
                        "access": 'download'
                    }
                ), kwargs["sanitized_GET"]
            )
        return render_to_string(
            "spider_filets/file.html", request=kwargs["request"],
            context=kwargs
        )

    def access_download(self, **kwargs):
        f = self.associated.attachedfiles.get(name="file")
        if getattr(settings, "FILE_DIRECT_DOWNLOAD", False):
            response = f.get_response()
        else:
            response = f.get_response(
                name=self.associated.name, request=kwargs["request"],
                add_extension=True
            )

        response["Access-Control-Allow-Origin"] = "*"
        return response

    def access_add(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Upload File"))
        return super().access_add(**kwargs)

    def access_update(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Update File"))
        return super().access_update(**kwargs)


@add_by_field(registry.contents, "_meta.model_name")
class TextFilet(LicenseMixin, DataContent):
    expose_name = "force"
    expose_description = True

    appearances = [{"name": "Text"}]

    class Meta:
        proxy = True

    def get_priority(self):
        # push to top
        if self.free_data.get("push"):
            return 1
        return 0

    def get_template_name(self, scope):
        # view update form
        if scope == "update_guest":
            return 'spider_base/edit_form.html'
        elif scope == "view":
            return 'spider_base/text.html'
        return super().get_template_name(scope)

    def get_content_description(self):
        # use javascript instead
        if self.prepared_attachements:
            t = self.prepared_attachements["attachedblobs"][0]
        else:
            t = self.associated.attachedblobs.filter(name="text").first()
        return " ".join(
            prepare_description(
                t.as_bytes.decode("utf8") if t else "", 51
            )[:50]
        )

    def get_info(self):
        ret = super().get_info()
        return "%sname=%s\x1efile_type=text\x1e" % (
            ret, self.associated.name

        )

    def get_form(self, scope):
        if scope in ("raw", "export", "list"):
            from .forms import RawTextForm as f
        else:
            from .forms import TextForm as f
        return f

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret["request"] = kwargs["request"]
        ret["source"] = kwargs.get("source", self.associated.usercomponent)
        ret["scope"] = kwargs["scope"]
        return ret

    def get_abilities(self, context):
        _abilities = set()
        source = context.get("source", self.associated.usercomponent)
        if source.id in self.free_data.get("editable_from", []):
            _abilities.add("update_guest")
        return _abilities

    def access_update_guest(self, **kwargs):
        kwargs["legend"] = \
            escape(_("Update \"%s\" (guest)") % self.__str__())
        kwargs["inner_form"] = False
        return self.access_update(**kwargs)

    def access_view(self, **kwargs):
        kwargs["object"] = self
        kwargs["text"] = \
            self.associated.attachedblobs.get(name="text").as_bytes.decode(
                "utf8"
            )
        return render_to_string(
            "spider_filets/text.html", request=kwargs["request"],
            context=kwargs
        )



__all__ = (
    "filter_components", "filter_contents", "listed_variants_q",
    "machine_variants_q", "active_protections_q",
    "info_and", "info_or", "travelprotection_types_q",
    "loggedin_active_tprotections_q"
)

from django.conf import settings
from django.db.models import Q

from spkcspider.constants import (
    ProtectionStateType, VariantType, loggedin_active_tprotections
)

_base_variants = ~(
    (
        ~Q(ctype__contains=VariantType.feature_connect) &
        Q(ctype__contains=VariantType.component_feature)
    ) |
    Q(ctype__contains=VariantType.content_feature)
)

machine_variants_q = (
    Q(ctype__contains=VariantType.machine) &
    _base_variants  # such features cannot be created
)

active_protections_q = (
    Q(state=ProtectionStateType.enabled) |
    Q(state=ProtectionStateType.instant_fail)
)

loggedin_active_tprotections_q = Q()
for i in loggedin_active_tprotections:
    loggedin_active_tprotections_q |= Q(
        info__contains="\x1etravel_protection_type={}\x1e".format(i)
    )

travelprotection_types_q = (
    Q(ctype__name="TravelProtection") |
    Q(ctype__name="SelfProtection")
)


listed_variants_q = (
    _base_variants &
    ~Q(ctype__contains=VariantType.unlisted)
)


def filter_components(
    search_filters, filter_unlisted=True, use_contents=True
):
    searchq = Q()
    searchq_exc = Q()
    notsearch = Q()

    counter = 0
    # against ddos
    max_counter = settings.SPIDER_MAX_SEARCH_PARAMETERS

    # list only unlisted if explicity requested or export is used
    # ComponentPublicIndex doesn't allow unlisted in any case
    # this is enforced by setting "is_special_user" to False
    if filter_unlisted:
        notsearch = ~Q(contents__info__contains="\x1eunlisted\x1e")

    for item in search_filters:
        if filter_unlisted and item == "_unlisted":
            continue
        if counter > max_counter:
            break
        counter += 1
        if len(item) == 0:
            continue
        use_strict = False
        if item.startswith("!!"):
            _item = item[1:]
        elif item.startswith("__"):
            _item = item[1:]
        elif item.startswith("!_"):
            _item = item[2:]
            use_strict = True
        elif item.startswith("!"):
            _item = item[1:]
        elif item.startswith("_"):
            _item = item[1:]
            use_strict = True
        else:
            _item = item
        qob = Q()
        if use_strict:
            if use_contents:
                qob |= Q(contents__info__contains="\x1e%s\x1e" % _item)
                # exclude unlisted from searchterms
                qob &= notsearch
        else:
            qob |= Q(description__icontains=_item)
        if _item == "index":
            qob |= Q(strength=10)
        elif use_strict:
            qob |= Q(
                name=_item,
                strength__lt=10
            )
        else:
            qob |= Q(
                name__icontains=_item,
                strength__lt=10
            )
        if item.startswith("!!"):
            searchq |= qob
        elif item.startswith("!"):
            searchq_exc |= qob
        else:
            searchq |= qob
    return (searchq & ~searchq_exc, counter)


def filter_contents(
    search_filters, ids=None, filter_unlisted=True, feature_exception=True,
    use_components=False
):
    """
    [summary]

    Arguments:
        search_filters {iterable} -- search filters

    Keyword Arguments:
        ids {[type]} -- [description] (default: {None})
        filter_unlisted {bool} -- [description] (default: {True})
        feature_exception {bool} -- [description] (default: {True})
        use_components {bool} -- [description] (default: {False})

    Returns:
        [type] -- [description]
    """
    searchq = Q()
    searchq_exc = Q()

    counter = 0
    unlisted_active = False
    # against ddos
    max_counter = settings.SPIDER_MAX_SEARCH_PARAMETERS

    for item in search_filters:
        if filter_unlisted is True and item == "_unlisted":
            continue
        elif item == "_unlisted":
            unlisted_active = True
        if counter > max_counter:
            break
        counter += 1
        if len(item) == 0:
            continue
        use_strict = False
        negate = False
        if item.startswith("!!"):
            _item = item[1:]
        elif item.startswith("__"):
            _item = item[1:]
        elif item.startswith("!_"):
            _item = item[2:]
            use_strict = True
            negate = True
        elif item.startswith("!"):
            _item = item[1:]
            negate = True
        elif item.startswith("_"):
            _item = item[1:]
            use_strict = True
        else:
            _item = item
        if use_strict:
            qob = Q(name=_item)
            qob |= Q(info__contains="\x1e%s\x1e" % _item)
            # can exclude/include specific usercomponents names
            if use_components:
                qob |= Q(usercomponent__name=_item)
        else:
            qob = Q(name__icontains=_item)
            qob |= Q(description__icontains=_item)
            qob |= Q(info__icontains=_item)
            # but don't be too broad with unspecific negation
            #   only apply on contents
            if use_components and not negate:
                qob |= Q(usercomponent__name__icontains=_item)
                qob |= Q(usercomponent__description__icontains=_item)
        if negate:
            searchq_exc |= qob
        else:
            searchq |= qob

    if ids:
        # id_collection contains int and str entries
        try:
            ids = map(lambda x: int(x), ids)
        except ValueError:
            # deny any access in case of an incorrect id
            ids = []

        searchq &= (
            Q(
                id__in=ids
            )
        )
    # if cannot set _unlisted
    if not unlisted_active:
        tmp = None
        if filter_unlisted is True:
            # if filter_unlisted is True generally exclude all
            tmp = Q(info__contains="\x1eunlisted\x1e")
        elif filter_unlisted is not False:
            # if filter_unlisted is int exclude all equal and below
            tmp = Q(
                info__contains="\x1eunlisted\x1e",
                priority__lte=filter_unlisted
            )
        if tmp:
            # except if features as they are shown always
            if feature_exception:
                tmp &= ~(
                    Q(
                        ctype__ctype__contains=VariantType.feature_connect
                    ) &
                    ~Q(
                        ctype__ctype__contains=VariantType.unlisted
                    )
                )
            searchq_exc |= tmp
    return (searchq & ~searchq_exc, counter)


def info_and(*args, **kwargs):
    q = Q()
    for query in args:
        if isinstance(query, Q):
            q &= query
        else:
            q &= Q(
                info__contains="\x1e{}\x1e".format(query)
            )
    for name, query in kwargs.items():
        if query is None:
            q &= Q(
                info__contains="\x1e{}=".format(name)
            )
        elif isinstance(query, (tuple, list)):
            for item in query:
                q &= Q(
                    info__contains="\x1e{}={}\x1e".format(name, item)
                )
        else:
            q &= Q(
                info__contains="\x1e{}={}\x1e".format(name, query)
            )
    return q


def info_or(*args, **kwargs):
    q = Q()
    for query in args:
        if isinstance(query, Q):
            q |= query
        else:
            q |= Q(
                info__contains="\x1e{}\x1e".format(query)
            )
    for name, query in kwargs.items():
        if query is None:
            q |= Q(
                info__contains="\x1e{}=".format(name)
            )
        elif isinstance(query, (tuple, list)):
            for item in query:
                q |= Q(
                    info__contains="\x1e{}={}\x1e".format(name, item)
                )
        else:
            q |= Q(
                info__contains="\x1e{}={}\x1e".format(name, query)
            )
    return q



# Generated by Django 3.0.1 on 2019-12-30 13:55

from django.db import migrations


def move_to_datacontent(apps, schema_editor):
    AssignedContent = apps.get_model("spider_base", "AssignedContent")
    AnchorKey = apps.get_model("spider_keys", "AnchorKey")
    AnchorServer = apps.get_model("spider_keys", "AnchorServer")
    PublicKey = apps.get_model("spider_keys", "PublicKey")
    DataContent = apps.get_model("spider_base", "DataContent")

    for a in AssignedContent.objects.filter(
        ctype__code=PublicKey._meta.model_name
    ):
        d = DataContent(associated=a)
        content = PublicKey.objects.get(id=a.object_id)
        d.save()
        a.attachedblobs.create(
            unique=True, name="key", blob=content.key.encode("utf8")
        )

    for a in AssignedContent.objects.filter(
        ctype__code=AnchorKey._meta.model_name
    ):
        d = DataContent(associated=a)
        content = AnchorKey.objects.get(id=a.object_id)
        d.quota_data["signature"] = content.signature
        d.save()

    for a in AssignedContent.objects.filter(
        ctype__code=AnchorServer._meta.model_name
    ):
        d = DataContent(associated=a)
        content = AnchorServer.objects.get(id=a.object_id)
        d.quota_data["new_url"] = content.new_url
        d.quota_data["old_urls"] = content.old_urls
        d.save()


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0011_auto_20191230_1305'),
        ('spider_keys', '0004_remove_anchorkey_key'),
    ]
    run_before = [
        ('spider_base', '0012_auto_20191230_1305'),
    ]

    operations = [
        migrations.RunPython(move_to_datacontent),
    ]



from django.contrib import admin

from .forms import TagLayoutAdminForm
from .models import TagLayout

# Register your models here.



@admin.register(TagLayout)
class TagLayoutAdmin(admin.ModelAdmin):
    form = TagLayoutAdminForm




__all__ = ["DataContentForm"]

from django import forms


class DataContentForm(forms.Form):
    instance = None
    # TODO: document
    free_fields = {}
    # TODO: document
    quota_fields = {}

    def __init__(self, instance, initial=None, **kwargs):
        if initial is None:
            initial = {}
        super().__init__(**kwargs)
        self.instance = instance
        self.initial.update(instance.free_data)
        self.initial.update(instance.quota_data)
        self.initial.update(initial)

    def get_prepared_attachements(self):
        return {}

    def clean(self):
        self.instance.full_clean()
        return super().clean()

    def save_m2m(self):
        # stub
        pass

    def save(self, commit=True):
        ret = self.instance
        ret.prepared_attachements = self.get_prepared_attachements()
        for key, default in self.free_fields.items():
            if key in self.fields:
                ret.free_data[key] = self.cleaned_data.get(key, default)
                if callable(ret.free_data[key]):
                    ret.free_data[key] = ret.free_data[key]()
        for key, default in self.quota_fields.items():
            if key in self.fields:
                ret.quota_data[key] = self.cleaned_data.get(key, default)
                if callable(ret.quota_data[key]):
                    ret.quota_data[key] = ret.quota_data[key]()
        if commit:
            ret.save()
        return ret



__all__ = [
    "VariantType", "TravelProtectionType",
    "dangerous_login_choices", "loggedin_active_tprotections",
    "essential_contents"
]

import enum

essential_contents = {
    "domainmode",
}


class VariantType(str, enum.Enum):
    # a not assigned
    # use persistent token
    #  can be used for some kind of federation
    #  warning: if token is deleted persisted content is deleted
    persist = "b"
    # update content is without form/for form updates it is not rendered
    # required for still beeing able to update elemental parameters
    raw_update = "c"
    # raw_add not required, archieved by returning response

    # don't list as contentvariant for user (for computer only stuff)
    # set unlisted as default in info field
    # if feature: unlist as a public visible feature
    unlisted = "d"
    # activates domain mode
    domain_mode = "e"
    # allow outside applications push content to spiders
    # adds by default unlisted info attribute
    # appears in features of userComponent
    # counts as foreign content
    # don't list as content variant for user (
    #    except feature_connect and not unique
    # )
    component_feature = "f"
    # same, but assigns to contents
    content_feature = "g"

    # is content unique for usercomponent
    # together with strength level 10: unique for user
    unique = "h"
    # can be used as anchor, will be hashed on verification if primary
    anchor = "i"
    # add as contentvariant (except if unlisted), and create feature
    feature_connect = "j"
    # add as machine creatable content
    machine = "k"
    # exclude from exports
    no_export = "l"

    def __str__(self):
        # output value instead of member name
        return self.value


class TravelProtectionType(str, enum.Enum):
    # don't show protected contents and components
    hide = "a"
    # switches to hiding if trigger was activated
    trigger_hide = "b"
    # disable login (and hides contents for logged in users)
    disable = "c"
    # disable login if triggered
    trigger_disable = "C"
    # wipe travel protected contents and components
    # Note: noticable if shared contents are removed
    wipe = "d"
    # wipe user, self destruct user on login, Note: maybe noticable
    wipe_user = "e"
    # trigger disable plus disable all components
    trigger_disable_user = "E"

    def __str__(self):
        # output value instead of member name
        return self.value

    @classmethod
    def as_choices(cls):
        from django.utils.translation import gettext_lazy as _
        return (
            (cls.hide, _("Hide")),
            (cls.trigger_hide, _("Hide if triggered")),
            (cls.disable, _("Disable login")),
            (cls.trigger_disable, _("Disable login if triggered")),
            (cls.wipe, _("Wipe")),
            (cls.wipe_user, _("Wipe User")),
            (
                cls.trigger_disable_user,
                _("Disable login and contents if triggered")
            ),
        )


dangerous_login_choices = {
    TravelProtectionType.wipe,
    TravelProtectionType.wipe_user
}

loggedin_active_tprotections = {
    TravelProtectionType.hide,
    TravelProtectionType.disable
}



__all__ = [
    "clean_graph", "get_hashob", "validate_request_default",
    "verify_tag_default", "domain_auth", "get_anchor_domain"
]

import functools
import logging
from urllib.parse import parse_qs, urlencode

import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from rdflib import XSD, Literal, URIRef

from django.conf import settings
from django.test import Client
from django.urls import reverse
from spkcspider.constants import spkcgraph

from .conf import get_requests_params


@functools.lru_cache(1)
def get_anchor_domain():
    # if None, set to default Site ID if models are ready
    _anchor_domain = getattr(
        settings, "VERIFIER_ANCHOR_DOMAIN", getattr(
            settings, "SPIDER_ANCHOR_DOMAIN", None
        )
    )
    if _anchor_domain:
        return _anchor_domain
    from django.contrib.sites.models import Site
    return Site.objects.get(id=settings.SITE_ID).domain


def validate_request_default(request, form):
    if not form.is_valid():
        return False

    return True


def verify_tag_default(tag, hostpart, ffrom) -> bool:
    """ return True if callback should be fired (if possible) """
    return tag.verification_state == "verified"


def domain_auth(source, hostpart):
    GET = parse_qs(source.get_params)
    if "token" not in GET:
        GET["token"] = "prefer"
    GET["intention"] = "domain"
    GET["referrer"] = "{}{}".format(
        hostpart,
        reverse("spider_verifier:create")
    )
    source.initialize_token()
    source.save(update_fields=["token"])
    GET["payload"] = urlencode(
        {
            "url": source.url,
            "update_secret": source.token
        }
    )

    url = "{}?{}".format(
        source.url, urlencode(GET, doseq=True)
    )
    ret = True
    params, inline_domain = get_requests_params(url)

    if inline_domain:
        response = Client().head(
            url, follow=True, secure=True, Connection="close",
            SERVER_NAME=inline_domain
        )
        if response.status_code >= 400:
            ret = False
    else:
        try:
            with requests.head(
                url, headers={"Connection": "close"},
                **params
            ) as resp:
                resp.close()
                resp.raise_for_status()
        except Exception:
            if settings.DEBUG:
                logging.exception("domain_auth failed")
            ret = False
    # other or own update_secret was successful or url has problems
    source.refresh_from_db()
    return ret


def clean_graph(graph, start, source, hostpart):
    mtype = str(graph.value(
        subject=start, predicate=spkcgraph["type"], any=not settings.DEBUG
    ))
    if not mtype:
        return None
    elif "UserComponent" == mtype:
        return "list"
    elif "SpiderTag" == mtype:
        if (
            URIRef(start),
            spkcgraph["abilities"],
            Literal("verify", datatype=XSD.string)
        ) in graph:
            return "layout_cb"
        ret = "layout"
        if domain_auth(source, hostpart):
            graph.set(
                (
                    start,
                    spkcgraph["action:view"],
                    Literal(source.get_url(), datatype=XSD.anyURI)
                )
            )
            ret = "layout_cb"
        return ret
    else:
        return "content"


def get_hashob():
    return hashes.Hash(
        getattr(
           settings, "VERIFICATION_HASH_ALGORITHM",
           settings.SPIDER_HASH_ALGORITHM
        ),
        backend=default_backend()
    )



__all__ = [
    "hex_size_of_bigid", "static_token_matcher", "host_tld_matcher",
    "TokenCreationError", "ProtectionResult", "ActionUrl", "protected_names"
]

import re
from collections import namedtuple

hex_size_of_bigid = 16


# user can change static token but elsewise the token stays static
static_token_matcher = re.compile(
    r"(?:[^?]*/|^)(?P<static_token>[^/?]+)/(?P<access>[^/?]+)"
)

host_tld_matcher = re.compile(
    r'^[^.]*?(?!\.)(?P<host>[^/?:]+?(?P<tld>\.[^/?:.]+)?)(?=[/?:]|$)(?!:/)'
)


class TokenCreationError(Exception):
    pass


ProtectionResult = namedtuple("ProtectionResult", ["result", "protection"])


ActionUrl = namedtuple("ActionUrl", ["name", "url"])
protected_names = {"index"}
